# built-in imports
import time
import json
from typing import Generator, Optional, Type
from collections import Counter

# local imports
from tritondse.seed              import Seed, SeedStatus
from tritondse.callbacks         import CallbackManager
from tritondse.coverage          import GlobalCoverage, CovItem, CoverageStrategy
from tritondse.worklist          import WorklistAddressToSet, FreshSeedPrioritizerWorklist, SeedScheduler
from tritondse.workspace         import Workspace
from tritondse.symbolic_executor import SymbolicExecutor
from tritondse.types             import SolverStatus, SymExType
import tritondse.logging

logger = tritondse.logging.get("seedmanager")


class SeedManager:
    """
    Seed Manager.
    This class is in charge of providing the next seed to execute by prioritizing
    them. It also holds various sets of pending seeds, corpus, crashes etc and
    manages them in the workspace.

    It contains basically 2 types of seeds which are:

    * pending seeds (kept in the seed scheduler). These are the seeds that might
      be selected to be run
    * seed consumed (corpus, crash, hangs) which are seeds not meant to be re-executed
      as they cannot lead to new paths, all candidate paths are UNSAT etc.
    """
    def __init__(self, coverage: GlobalCoverage, workspace: Workspace, smt_queries_limit: int, callback_manager: CallbackManager = None, seed_scheduler_class: Type[SeedScheduler] = None):
        """
        :param coverage: global coverage object. The instance will be updated by the seed manager
        :type coverage: GlobalCoverage
        :param workspace: workspace instance object.
        :type workspace: Workspace
        :param smt_queries_limit: maximum number of queries for  a given execution
        :type smt_queries_limit: int
        :param scheduler: seed scheduler object to use as scheduling strategy
        :type scheduler: SeedScheduler
        """
        self.smt_queries_limit = smt_queries_limit
        self.workspace = workspace
        self.coverage = coverage
        if seed_scheduler_class is None:
            self.worklist = FreshSeedPrioritizerWorklist(self)
        else:
            self.worklist = seed_scheduler_class(self)
        self.cbm = callback_manager

        self.corpus = set()
        self.crash = set()
        self.hangs = set()
        self.fails = set()

        self.__load_seed_workspace()

        self._solv_count = 0
        self._solv_time_sum = 0
        self._solv_status = {SolverStatus.SAT: 0, SolverStatus.UNSAT: 0, SolverStatus.UNKNOWN: 0, SolverStatus.TIMEOUT: 0}
        self._stat_branch_reverted = Counter()
        self._stat_branch_fail = Counter()
        self._current_solv_time = 0

    @property
    def total_solving_time(self) -> float:
        return self._solv_time_sum

    def __load_seed_workspace(self):
        """ Load seed from the workspace """
        # Load seed from the corpus
        for seed in self.workspace.iter_corpus():
            self.corpus.add(seed)
        # Load hangs
        for seed in self.workspace.iter_hangs():
            self.hangs.add(seed)
        # Load crashes
        for seed in self.workspace.iter_crashes():
            self.crash.add(seed)
        # Load worklist
        for seed in self.workspace.iter_worklist():
            self.worklist.add(seed)
        # Load fails
        for seed in self.workspace.iter_fails():
            self.fails.add(seed)

    def is_new_seed(self, seed: Seed) -> bool:
        """
        Check if a seed is a new one (not into corpus, crash, hangs and fails)

        :param seed: The seed to test
        :type seed: Seed
        :return: True if the seed is a new one

        .. warning:: That function does not check that the seed
                     is not in the pending seeds queue
        """
        return sum(seed in x for x in [self.corpus, self.crash, self.hangs, self.fails]) == 0

    def add_seed_queue(self, seed: Seed) -> None:
        """
        Add a seed to to appropriate internal queue depending
        on its status. If it is new it is added in pending seed,
        if OK, HANG or CRASH it the appropriate set.
        **Note that the seed is not written in the workspace**

        :param seed: Seed to add in an internal queue
        :type seed: Seed
        """
        # Add the seed to the appropriate list
        if seed.status == SeedStatus.NEW:
            self.worklist.add(seed)
        elif seed.status == SeedStatus.OK_DONE:
            self.corpus.add(seed)
        elif seed.status == SeedStatus.HANG:
            self.hangs.add(seed)
        elif seed.status == SeedStatus.CRASH:
            self.crash.add(seed)
        elif seed.status == SeedStatus.FAIL:
            self.fails.add(seed)
        else:
            assert False

    def post_execution(self, execution: SymbolicExecutor, seed: Seed, solve_new_path: int = True) -> float:
        """
        Function called after each execution. It updates the global
        code coverage object, and tries to generate new paths through
        SMT in accordance with the seed scheduling strategy.

        :param execution: The current execution
        :type execution: SymbolicExecutor
        :param seed: The seed of the execution
        :type seed: Seed
        :param solve_new_path: Whether or not to solve constraint to find new paths
        :type solve_new_path: bool
        :return: Total SMT solving time
        """

        # Update instructions covered from the last execution into our exploration coverage
        self.coverage.merge(execution.coverage)
        self.worklist.update_worklist(execution.coverage)

        # if the seed have target checks that we covered it
        if seed.target:
            color = ("YES", 92) if execution.coverage.is_covered(seed.target) else ("NO", 91)
            logger.info(f"Seed covered its target: \033[{color[1]}m{color[0]}\033[0m")

        # reset the current solving time
        self._current_solv_time = 0

        # Iterate all pending seeds to be added in the right location
        for s in execution.pending_seeds:
            if self.is_new_seed(s):
                if not s.coverage_objectives:      # If they don't have objectives set the Ellipsis wildcard
                    s.coverage_objectives.add(...)
                self._add_seed(s)  # will add the seed in both internal queues & workspace
            else:
                logger.warning(f"dropping enqueued seed: {s.hash} (already seen)")

        # Update the current seed queue
        if seed.status == SeedStatus.NEW:
            logger.error(f"seed not meant to be NEW at the end of execution ({seed.hash}) (dropped)")
            self.drop_seed(seed)

        elif seed.status in [SeedStatus.HANG, SeedStatus.CRASH, SeedStatus.FAIL]:
            self.archive_seed(seed)
            # NOTE: Do not perform further processing on the seed (like generating inputs from it)

        elif seed.status == SeedStatus.OK_DONE:
            if self.coverage.can_improve_coverage(execution.coverage) or self.coverage.can_cover_symbolic_pointers(execution):
                items = self.coverage.new_items_to_cover(execution.coverage)
                seed.coverage_objectives = items  # Set its new objectives

                if self.worklist.can_solve_models() and solve_new_path:     # No fresh seeds pending thus can solve model
                    logger.info(f'Seed {seed.hash} generate new coverage')
                    self._generate_new_inputs(execution)
                    self.archive_seed(seed)
                else:
                    logger.info(f"Seed {seed.hash} push back in worklist (to unstack fresh)")
                    seed.status = SeedStatus.NEW  # Reset its status for later run
                    self.add_seed_queue(seed)  # will be pushed back in worklist
            else:
                self.archive_seed(seed)
                logger.info(f'Seed {seed.hash} archived cannot generate new coverage [{seed.status.name}]')

        else:
            assert False

        logger.info(f"Corpus:{len(self.corpus)} Crash:{len(self.crash)}")
        self.worklist.post_execution()
        logger.info(f"Coverage instruction:{self.coverage.unique_instruction_covered} covitem:{self.coverage.unique_covitem_covered}")
        return self._current_solv_time

    def _generate_new_inputs(self, execution: SymbolicExecutor):
        # Generate new inputs
        for new_input in self.__iter_new_inputs(execution):
            # Check if we already have executed this new seed
            if self.is_new_seed(new_input):
                self.worklist.add(new_input)
                self.workspace.save_seed(new_input)
                logger.info(f'New seed model {new_input.filename} dumped [{new_input.status.name}]')
            else:
                logger.info(f"New seed {new_input.filename} has already been generated")

    def __iter_new_inputs(self, execution: SymbolicExecutor) -> Generator[Seed, None, None]:
        # Get the astContext
        actx = execution.pstate.actx

        # We start with any input. T (Top)
        path_predicate = [actx.equal(actx.bvtrue(), actx.bvtrue())]

        # Define a limit of branch constraints
        smt_queries = 0

        # Solver status
        status = None
        path_constraints = execution.pstate.get_path_constraints()
        total_len = len(path_constraints)
        path_generator = self.coverage.iter_new_paths(path_constraints)

        try:
            while True:
                # If smt_queries_limit is zero: unlimited queries
                # If smt_queries_limit is negative: no query
                if self.smt_queries_limit < 0:
                    logger.info(f'The configuration is defined as: no query')
                    break

                typ, p_prefix, branch, covitem, ith = path_generator.send(status)

                # Create edge in case of conditional branch, for all the other the edge shall be already set
                edge = (branch['srcAddr'], branch['dstAddr']) if typ == SymExType.CONDITIONAL_JMP else covitem

                # Add path_prefix in path predicate (regardless on whether we solve the item or not)
                path_predicate.extend(x.getTakenPredicate() for x in p_prefix)

                expr = branch['constraint'] if typ == SymExType.CONDITIONAL_JMP else branch

                # Call on_branch_solving, if one replies False does not solve the branch
                if self.cbm is not None:
                    cb_result = all(cb(execution, execution.pstate, edge, typ, expr, path_predicate) for cb in self.cbm.get_on_solving_callback())
                else:
                    cb_result = True

                # Skip processing the current path in case the result of the
                # callbacks return False.
                if not cb_result:
                    status = None
                    continue

                # Create the constraint
                if typ in [SymExType.DYNAMIC_JMP, SymExType.SYMBOLIC_READ, SymExType.SYMBOLIC_WRITE]:
                    expr, (addr, tgt) = branch, covitem   # branch and covitem have a different meaning here
                    ts = time.time()
                    results = execution.pstate.solve_enumerate_expression(expr, path_predicate, [tgt], execution.config.smt_enumeration_limit)  # enumerate values

                    # all stats updates
                    solve_time = time.time() - ts
                    count = len(results)
                    status = SolverStatus.SAT if count else SolverStatus.UNSAT
                    self._update_solve_stats(None, status, solve_time, count)

                    results = [(x[0], (addr, x[1])) for x in results]   # extract results
                    logger.info(f'pc:{ith}/{total_len} | Query n°{smt_queries}-{smt_queries+count}, enumerate:{expr} (time: {solve_time:.02f}s) values:[{count}:{self._pp_smt_status(status)}]')
                    smt_queries += count+1  # for the unsat

                elif typ == SymExType.CONDITIONAL_JMP:
                    # if debug_pp=True solve the branch that has been taken
                    branch_cst = actx.lnot(branch['constraint']) if execution.debug_pp else branch['constraint']
                    constraint = actx.land(path_predicate + [branch_cst])

                    # Solve the constraint
                    ts = time.time()
                    status, model = execution.pstate.solve(constraint, with_pp=False)  # Do not use path predicate as we are iterating it
                    solve_time = time.time() - ts
                    self._update_solve_stats(covitem, status, solve_time)
                    results = [(model, covitem)]
                    smt_queries += 1
                    logger.info(f'pc:{ith}/{total_len} | Query n°{smt_queries}, solve:{self.coverage.pp_item(covitem)} (time: {solve_time:.02f}s) [{self._pp_smt_status(status)}]')
                else:
                    assert False

                if status == SolverStatus.SAT:
                    for model, covitem in results:
                        new_seed = execution.mk_new_seed_from_model(model)
                        # Trick to keep track of which target a seed is meant to cover
                        new_seed.coverage_objectives.add(covitem)
                        new_seed.meta_fname.append(self.pp_meta_filename(covitem, typ))
                        new_seed.target = covitem if typ == SymExType.CONDITIONAL_JMP else None
                        yield new_seed  # Yield the seed to get it added in the worklist
                else:
                    pass

                # Check if we reached the limit of query
                if self.smt_queries_limit and smt_queries >= self.smt_queries_limit:
                    logger.info(f'Limit of query reached. Stop asking for models')
                    break

        except StopIteration:  # We have iterated the whole path generator
            pass

    def _update_solve_stats(self, covitem: Optional[CovItem], status: SolverStatus, solving_time: float, count=1):
        self._solv_count += count
        self._solv_time_sum += solving_time
        self._current_solv_time += solving_time
        self._solv_status[status] += count
        logger.debug(f'Solve stats: solve_count={self._solv_count} solving_time={solving_time} solve_time_sum={self._solv_time_sum} current_solve_time={self._current_solv_time} solv_status={status} / {self._solv_status[status]}')

        if covitem:
            if status == SolverStatus.SAT:
                self._stat_branch_reverted[covitem] += count  # Update stats
                if covitem in self._stat_branch_fail:
                    self._stat_branch_fail.pop(covitem)
            elif status == SolverStatus.UNSAT:
                self._stat_branch_fail[covitem] += count

    def pick_seed(self) -> Optional[Seed]:
        """
        Get the next seed to be executed by querying it
        in the seed scheduler.

        :returns: Seed to execute from the pending seeds
        :rtype: Seed
        """
        return self.worklist.pick()

    def seeds_available(self) -> bool:
        """
        Checks whether or not there is still pending seeds to process.

        :returns: True if seeds are still pending
        """
        return self.worklist.has_seed_remaining()

    def add_new_seed(self, seed: Seed) -> None:
        """
        Add the given seed in the manager.
        The function uses its type to know where to add the seed.

        :param seed: seed to add
        :type seed: Seed
        """
        if self.is_new_seed(seed):
            self._add_seed(seed)
            logger.debug(f'Seed {seed.filename} dumped [{seed.status.name}]')
        else:
            logger.debug(f"seed {seed} is already known (not adding it)")

    def _add_seed(self, seed: Seed) -> None:
        """ Add the seed in both internal queues but also workspace """
        self.add_seed_queue(seed)
        self.workspace.save_seed(seed)

    def drop_seed(self, seed: Seed) -> None:
        """
        Drop a seed that is not of interest anymore.
        The function thus switch its status to ``OK_DONE``
        and move it in the corpus. *(the seed is not removed
        from the corpus)*

        :param seed: seed object to drop
        :type seed: Seed
        """
        self.archive_seed(seed)

    def archive_seed(self, seed: Seed, status: SeedStatus = None) -> None:
        """
        Send a seed in the corpus. As such, the seed
        is not meant to be used anymore (for finding
        new seeds).

        :param seed: seed object
        :type seed: Seed
        :param status: optional status to assign the seed
        :type status: SeedStatus
        """
        if status:
            seed.status = status
        self.add_seed_queue(seed)  # Will put it in the corpus
        self.workspace.update_seed_location(seed)  # Will put it in the corpus in files

    def post_exploration(self) -> None:
        """
        Function called at the end of exploration. It perform
        some stats printing, but would also perform any clean
        up tasks. *(not meant to be called by the user)*
        """
        # Do things you would do at the very end of exploration
        # (or just before it becomes idle)
        stats = {
            "total_solving_time": self._solv_time_sum,
            "total_solving_attempt": self._solv_count,
            "branch_reverted": {str(k): v for k, v in self._stat_branch_reverted.items()}, # convert covitem to str whatever it is
            "branch_not_solved": {str(k): v for k, v in self._stat_branch_fail.items()},  # convert covitem to str whatever it is
            "UNSAT": self._solv_status[SolverStatus.UNSAT],
            "SAT": self._solv_status[SolverStatus.SAT],
            "TIMEOUT": self._solv_status[SolverStatus.TIMEOUT]
        }
        self.workspace.save_metadata_file("solving_stats.json", json.dumps(stats, indent=2))
        logger.info(f"Branches reverted: {len(self._stat_branch_reverted)}  Branches still fail: {len(self._stat_branch_fail)}")
        self.worklist.post_exploration(self.workspace)

    @staticmethod
    def _pp_smt_status(status: SolverStatus) -> str:
        """ The pretty print function of the solver status """
        mapper = {SolverStatus.SAT: 92, SolverStatus.UNSAT: 91, SolverStatus.TIMEOUT: 93, SolverStatus.UNKNOWN: 95}
        return f"\033[{mapper[status]}m{status.name}\033[0m"

    def pp_meta_filename(self, covitem: CovItem, typ: SymExType) -> str:
        """pretty-print a covitem"""
        if typ == SymExType.CONDITIONAL_JMP:
            pp_item = self.coverage.pp_item(covitem)
        else:
            pp_item = f"({covitem[0]:#08x}:@[{covitem[1]:#08x}])"
        map = {SymExType.CONDITIONAL_JMP: "CC",
               SymExType.SYMBOLIC_READ: "SR",
               SymExType.SYMBOLIC_WRITE: "SW",
               SymExType.DYNAMIC_JMP: "DYN"}
        return f"{map[typ]}_{pp_item}"



# built-in imports
import json
import os
import subprocess
from pathlib import Path
import time
from typing import List, Optional, Union
from collections import Counter


# local imports
import tritondse # NOTE We need this import so we can use it to determine the path of this file.
from tritondse import Config, Program, SymbolicExecutor, CoverageStrategy, CoverageSingleRun
import tritondse.logging

logger = tritondse.logging.get("tracer")


class TraceException(Exception):
    pass


class Trace:
    def __init__(self):
        pass

    @staticmethod
    def run(strategy: CoverageStrategy, binary_path: str, args: List[str], output_path: str, dump_trace: bool = False, stdin_file=None) -> bool:
        """Run the binary passed as argument and return the coverage.

        :param strategy: Coverage strategy.
        :type strategy: :py:obj:`CoverageStrategy`.
        :param binary_path: Path to the binary.
        :type binary_path: :py:obj:`str`.
        :param args: List of arguments to pass to the binary.
        :type args: :py:obj:`List[str]`.
        :type output_path: File where to store trace
        :param dump_trace: Enable gather the trace
        :param stdin_file: Path to the file that will act as stdin.
        :type args: :py:obj:`str`.
        """
        raise NotImplementedError()

    @property
    def trace(self) -> List[int]:
        raise NotImplementedError()

    @staticmethod
    def from_file(file: Union[str, Path]) -> 'QBDITrace':
        raise NotImplementedError()

    @property
    def coverage(self) -> CoverageSingleRun:
        """
        Coverage generated by the trace

        :return: CoverageSingleRun object
        """
        raise NotImplementedError()

    def get_coverage(self) -> CoverageSingleRun:
        """Return the execution coverage.

        :return: :py:obj:`CoverageSingleRun`.
        """
        return self.coverage

    @property
    def strategy(self) -> CoverageStrategy:
        """
        Return the coverage strategy with which this trace
        was generated with.

        :return: :py:obj:`CoverageStrategy`
        """
        return self.coverage.strategy


class TritonTrace(Trace):

    def __init__(self):
        super().__init__()

        self._coverage = None

    @staticmethod
    def run(strategy: CoverageStrategy, binary_path: str, args: List[str], output_path: str, dump_trace: bool = False, stdin_file=None) -> bool:
        # Override stdin with the input file.
        if stdin_file:
            os.dup2(os.open(stdin_file, os.O_RDONLY), 0)

        config = Config(coverage_strategy=strategy)

        se = SymbolicExecutor(config)

        se.load(Program(binary_path))

        se.run()

        trace = TritonTrace()
        trace._coverage = se.coverage
        # FIXME: Writing the coverage to a file

    @staticmethod
    def from_file(file: Union[str, Path]) -> 'QBDITrace':
        # FIXME: Reading coverage file from a file
        return trace

    @property
    def coverage(self) -> CoverageSingleRun:
        return self._coverage


class QBDITrace(Trace):

    QBDI_SCRIPT_FILEPATH = Path(tritondse.__file__).parent / 'qbdi_trace.py'

    def __init__(self):
        super().__init__()
        self._coverage = None
        self._trace = None
        self.modules = {}

    @staticmethod
    def run(strategy: CoverageStrategy, binary_path: str, args: List[str], output_path: str, dump_trace: bool = False, stdin_file=None, timeout=None, cwd=None) -> bool:
        if not Path(binary_path).exists():
            raise FileNotFoundError()

        if stdin_file and not Path(stdin_file).exists():
            raise FileNotFoundError()

        args = [] if not args else args

        cmdlne = f'timeout {timeout} python -m pyqbdipreload {QBDITrace.QBDI_SCRIPT_FILEPATH}'.split(' ') + [binary_path] + args
        cmdlne = " ".join(cmdlne)

        logger.debug(f'Command line: {cmdlne}')

        # Set environment variables.
        environ = {
            'PYQBDIPRELOAD_COVERAGE_STRATEGY': strategy.name,
            'PYQBDIPRELOAD_OUTPUT_FILEPATH': output_path,
            'PYQBDIPRELOAD_DUMP_TRACE': str(dump_trace),
            'LD_BIND_NOW': '1',
        }
        environ.update(os.environ)

        # Open stdin file if it is present.
        stdin_fp = open(stdin_file, 'rb') if stdin_file else None

        # Run QBDI tool.
        process = subprocess.Popen(cmdlne,
                                   shell=True,
                                   stdin=stdin_fp,
                                   stdout=subprocess.DEVNULL,
                                   stderr=subprocess.DEVNULL,
                                   cwd=cwd,
                                   env=environ)
        try:
            stdout, stderr = process.communicate(timeout=timeout)
            # for line in stdout.split(b"\n"):
            #     logger.debug(f"stdout: {line}")
            # for line in stderr.split(b"\n"):
            #     logger.debug(f"stdout: {line}")
        except subprocess.TimeoutExpired:
            process.wait()
            # logger.warning('QBDI tracer timeout expired!')
            raise TraceException('QBDI tracer timeout expired')

        if stdin_fp:
            stdin_fp.close()

        return Path(output_path).exists()

    @staticmethod
    def from_file(coverage_path: str) -> 'QBDITrace':
        """Load coverage from a file.

        :param coverage_path: Path to the coverage file.
        :type coverage_path: :py:obj:`str`.
        """
        trace = QBDITrace()

        logger.debug(f'Loading coverage file: {coverage_path}')
        with open(coverage_path, 'rb') as fd:
            data = json.load(fd)

        cov = CoverageSingleRun(CoverageStrategy[data["coverage_strategy"]])
        cov.covered_instructions = Counter({int(k): v for k, v in data["covered_instructions"].items()})

        for (src, dst, not_taken) in data["covered_items"]:
            if not_taken is None:
                cov.add_covered_dynamic_branch(src, dst)
            else:
                cov.add_covered_branch(src, dst, not_taken)

        trace._coverage = cov
        trace._trace = data['trace']
        trace.modules = data['modules_base']

        return trace

    @property
    def coverage(self) -> CoverageSingleRun:
        """
        CoverageSingleRun associated with the trace.

        :return: coverage object
        """
        if not self._coverage:
            logger.warning("Please .run() the trace before querying coverage")

        return self._coverage

    @property
    def trace(self) -> List[int]:
        """
        List of addresses executed.

        :return: list of addresses
        """
        return self._trace


if __name__ == "__main__":
    import sys
    if len(sys.argv) < 2:
        print("Usage: trace.py program [args]")
        sys.exit(1)

    tritondse.logging.enable()

    if QBDITrace.run(CoverageStrategy.EDGE, sys.argv[1], sys.argv[2:], "/tmp/test.cov", dump_trace=False):
        coverage = QBDITrace.from_file("/tmp/test.cov")
    else:
        print("Something went wrong during trace generation")



# built-in imports
from __future__ import annotations
from enum import Enum, auto
from typing import Callable, Tuple, List, Optional, Union, Any
import enum_tools.documentation
# third-party imports
from triton import CALLBACK, Instruction, MemoryAccess, OPCODE

# local imports
from tritondse.process_state import ProcessState
from tritondse.types import Addr, Input, Register, Expression, Edge, SymExType, AstNode
from tritondse.thread_context import ThreadContext
from tritondse.seed import Seed
from tritondse.memory import MemoryAccessViolation
import tritondse.logging

logger = tritondse.logging.get()  # get root tritondse logger


@enum_tools.documentation.document_enum
class CbPos(Enum):
    """ Enmus representing callback position """
    BEFORE = auto()  # doc: Callback should be executed before the hook location interpretation
    AFTER = auto()   # doc: Callback should be executed after the hook location interpreation


@enum_tools.documentation.document_enum
class CbType(Enum):
    """ Enmus representing all kind of callbacks """
    CTX_SWITCH = auto()     # doc: context-switch callback
    MEMORY_READ = auto()    # doc: memory read callback
    MEMORY_WRITE = auto()   # doc: memory write callback
    POST_RTN = auto()       # doc: routine callback (after)
    POST_ADDR = auto()      # doc: address class back (after)
    POST_EXEC = auto()      # doc: post execution callback, after a SymbolicExecutor run
    POST_INST = auto()      # doc: post instruction callback (all of them)
    PRE_ADDR = auto()       # doc: pre address callback
    PRE_EXEC = auto()       # doc: pre-execution callback (before launching a Symbolic Executor)
    PRE_INST = auto()       # doc: pre instruction callback (all of them)
    PRE_RTN = auto()        # doc: routine callback (before)
    REG_READ = auto()       # doc: a register is read
    REG_WRITE = auto()      # doc: a register is written
    NEW_INPUT = auto()      # doc: new input has been generated
    EXPLORE_STEP = auto()   # doc: an exploration step is reached. Called in between each SymbolicExecutor run
    PRE_MNEM = auto()       # doc: callback on a specific mnemonic (as a string) (before)
    POST_MNEM = auto()      # doc: callback on a specific mnemonic (as a string) (after)
    PRE_OPCODE = auto()     # doc: callback on a specific opcode (as bytes) (before)
    POST_OPCODE = auto()    # doc: callback on a specific opcode (as bytes) (after)
    BRANCH_COV = auto()     # doc: a new branch is getting covered
    SYMEX_SOLVING = auto()  # doc: An SMT formula is getting solved
    MEM_VIOLATION = auto()  # doc: A memory violation occured


AddrCallback            = Callable[['SymbolicExecutor', ProcessState, Addr], None]
ExplorationStepCallback = Callable[['SymbolicExplorator'], None]
InstrCallback           = Callable[['SymbolicExecutor', ProcessState, Instruction], None]
MemReadCallback         = Callable[['SymbolicExecutor', ProcessState, MemoryAccess], None]
MemWriteCallback        = Callable[['SymbolicExecutor', ProcessState, MemoryAccess, int], None]
MnemonicCallback        = Callable[['SymbolicExecutor', ProcessState, OPCODE], None]
SymExSolvingCallback    = Callable[['SymbolicExecutor', ProcessState, Edge, SymExType, AstNode, List[AstNode]], bool]
BranchCoveredCallback   = Callable[['SymbolicExecutor', ProcessState, Edge], bool]
NewInputCallback        = Callable[['SymbolicExecutor', ProcessState, Seed], Optional[Seed]]
OpcodeCallback          = Callable[['SymbolicExecutor', ProcessState, bytes], None]
RegReadCallback         = Callable[['SymbolicExecutor', ProcessState, Register], None]
RegWriteCallback        = Callable[['SymbolicExecutor', ProcessState, Register, int], None]
RtnCallback             = Callable[['SymbolicExecutor', ProcessState, str, Addr], Optional[Union[int, Expression]]]
SymExCallback           = Callable[['SymbolicExecutor', ProcessState], None]
ThreadCallback          = Callable[['SymbolicExecutor', ProcessState, ThreadContext], None]
MemoryViolationCallback = Callable[['SymbolicExecutor', ProcessState, MemoryAccessViolation], None]


class ProbeInterface(object):
    """ The Probe interface """
    def __init__(self):
        self._cbs: List[Tuple[CbType, Callable, Optional[str]]] = []  #: list of callback infos

    @property
    def callbacks(self) -> List[Tuple[CbType, Callable, Optional[Any]]]:
        return self._cbs

    def _add_callback(self, typ: CbType, callback: Callable, arg: str = None):
        """ Add a callback """
        self._cbs.append((typ, callback, arg))


class CallbackManager(object):
    """
    Class used to aggregate all callbacks that can be plugged
    inside a SymbolicExecutor running session. The internal
    structure ensure that check the presence of callback can
    be made in Log(N). All callbacks are designed to be read-only
    """

    def __init__(self):
        self._se = None

        # SymbolicExplorator callbacks
        self._step_cbs = []  # Callback called between each exploration steps

        # SymbolicExecutor callbacks
        self._pc_addr_cbs        = {}  # addresses reached
        self._opcode_cbs         = {}  # opcode before and after
        self._mnemonic_cbs       = {}  # mnemonic before and after
        self._instr_cbs          = {CbPos.BEFORE: [], CbPos.AFTER: []}  # all instructions
        self._pre_exec           = []  # before execution
        self._post_exec          = []  # after execution
        self._ctx_switch         = []  # on each thread context switch (implementing pre/post?)
        self._new_input_cbs      = []  # each time an SMT model is get
        self._branch_solving_cbs = []  # each time a branch is about to be solved
        self._branch_covered_cbs = []  # each time a branch is covered
        self._pre_rtn_cbs        = {}  # before imported routine calls ({str: [RtnCallback]})
        self._post_rtn_cbs       = {}  # after imported routine calls ({str: [RtnCallback]})
        self._mem_violation_cbs  = []  # called when an exception is raised

        # Triton callbacks
        self._mem_read_cbs  = []  # memory reads
        self._mem_write_cbs = []  # memory writes
        self._reg_read_cbs  = []  # register reads
        self._reg_write_cbs = []  # register writes
        self._empty         = True

        # Temporary mapping of function name to register
        self._func_to_register = {}

    def is_empty(self) -> bool:
        """
        Check whether a callback has alreday been registered or not

        :return: True if no callback were registered
        """
        return self._empty


    def is_binded(self) -> bool:
        """
        Check if the callback manager has already been binded on a given process state.

        :return: True if callbacks are binded on a process state
        """
        return bool(self._se) # and self._se.uid == se.uid)


    def _trampoline_mem_read_cb(self, ctx, mem):
        """
        This function is the trampoline callback on memory read from triton to tritondse

        :param ctx: TritonContext
        :param mem: MemoryAccess
        :return: None
        """
        if self._se.pstate.memory.callbacks_enabled():
            for cb in self._mem_read_cbs:
                cb(self._se, self._se.pstate, mem)


    def _trampoline_mem_write_cb(self, ctx, mem, value):
        """
        This function is the trampoline callback on memory write from triton to tritondse

        :param ctx: TritonContext
        :param mem: MemoryAccess
        :param value: int
        :return: None
        """
        if self._se.pstate.memory.callbacks_enabled():
            for cb in self._mem_write_cbs:
                cb(self._se, self._se.pstate, mem, value)


    def _trampoline_reg_read_cb(self, ctx, reg):
        """
        This function is the trampoline callback on register read from triton to tritondse

        :param ctx: TritonContext
        :param reg: Register
        :return: None
        """
        for cb in self._reg_read_cbs:
            cb(self._se, self._se.pstate, reg)


    def _trampoline_reg_write_cb(self, ctx, reg, value):
        """
        This function is the trampoline callback on register write from triton to tritondse

        :param ctx: TritonContext
        :param reg: Register
        :param value: int
        :return: None
        """
        for cb in self._reg_write_cbs:
            cb(self._se, self._se.pstate, reg, value)

    def unbind(self) -> None:
        """
        Unbind callbacks from the current process state as well as from
        the Triton Context object.
        """
        if self.is_binded():
            self._se.pstate.clear_triton_callbacks()
            self._se = None

    def bind_to(self, se: 'SymbolicExecutor') -> None:
        """
        Bind callbacks on the given process state. That step is required
        to register callbacks on the Triton Context object. This is also
        used to keep a reference on the SymbolicExecutor object;

        IMPORTANT You MUST call `unbind` once you finish using the
        SymbolicExecutor.

        :param se: SymbolicExecutor on which to bind callbacks
        :type se: SymbolicExecutor
        """
        if self.is_binded() and self._se != se:
            logger.warning("Callback_manager already binded (on a different executor instance)")
        # assert not self.is_binded()

        # NOTE This creates a circular dependency between the SymbolicExecutor
        #      received and this object, as the SymbolicExecutor keeps a
        #      reference to it. Therefore, it is necessary to call `unbind`
        #      once you finish using the executor.
        self._se = se

        # Register only one trampoline by kind of callback. It will be the role
        # of the trampoline to call every registered tritondse callbacks.

        if self._mem_read_cbs:
            se.pstate.register_triton_callback(CALLBACK.GET_CONCRETE_MEMORY_VALUE, self._trampoline_mem_read_cb)

        if self._mem_write_cbs:
            se.pstate.register_triton_callback(CALLBACK.SET_CONCRETE_MEMORY_VALUE, self._trampoline_mem_write_cb)

        if self._reg_read_cbs:
            se.pstate.register_triton_callback(CALLBACK.GET_CONCRETE_REGISTER_VALUE, self._trampoline_reg_read_cb)

        if self._reg_write_cbs:
            se.pstate.register_triton_callback(CALLBACK.SET_CONCRETE_REGISTER_VALUE, self._trampoline_reg_write_cb)

        # Check if there is a program on which to register functions callback
        if self._func_to_register:
            if se.loader:
                for fname in list(self._func_to_register):
                    cbs = self._func_to_register.pop(fname)
                    addr = se.loader.find_function_addr(fname)
                    if addr:
                        for cb in cbs:
                            self.register_pre_addr_callback(addr, cb)
                    else:
                        logger.warning(f"can't find function '{fname}' in {se.loader}")
            else:
                logger.warning(f"function callback to resolve but no program provided")


    def register_addr_callback(self, pos: CbPos, addr: Addr, callback: AddrCallback) -> None:
        """
        Register a callback function on a given address before or after the execution
        of the associated instruction.

        :param pos: When to trigger the callback (before or after) execution of the instruction
        :type pos: CbPos
        :param addr: Address where to trigger the callback
        :type addr: :py:obj:`tritondse.types.Addr`
        :param callback: callback function
        :type callback: :py:obj:`tritondse.callbacks.AddrCallback`
        """
        if addr not in self._pc_addr_cbs:
            self._pc_addr_cbs[addr] = {CbPos.BEFORE: [], CbPos.AFTER: []}

        self._pc_addr_cbs[addr][pos].append(callback)
        self._empty = False


    def register_pre_addr_callback(self, addr: Addr, callback: AddrCallback) -> None:
        """
        Register pre address callback

        :param addr: Address where to trigger the callback
        :type addr: :py:obj:`tritondse.types.Addr`
        :param callback: callback function
        :type callback: :py:obj:`tritondse.callbacks.AddrCallback`
        """
        self.register_addr_callback(CbPos.BEFORE, addr, callback)


    def register_post_addr_callback(self, addr: Addr, callback: AddrCallback) -> None:
        """
        Register post-address callback. Equivalent to register a pre-address on the
        return site. *(assume the function returns)*

        :param addr: Address where to trigger the callback
        :type addr: :py:obj:`tritondse.types.Addr`
        :param callback: callback function
        :type callback: :py:obj:`tritondse.callbacks.AddrCallback`
        """
        self.register_addr_callback(CbPos.AFTER, addr, callback)


    def get_address_callbacks(self, addr: Addr) -> Tuple[List[AddrCallback], List[AddrCallback]]:
        """
        Get all the pre/post callbacks for a given address.

        :param addr: Address where to trigger the callback
        :type addr: :py:obj:`tritondse.types.Addr`
        :return: tuple of lists containing callback functions for pre/post respectively
        """
        cbs = self._pc_addr_cbs.get(addr, None)
        if cbs:
            return cbs[CbPos.BEFORE], cbs[CbPos.AFTER]
        else:
            return [], []


    def register_opcode_callback(self, pos: CbPos, opcode: bytes, callback: OpcodeCallback) -> None:
        """
        Register a callback function on a given opcode before or after the execution
        of the associated instruction.

        :param pos: When to trigger the callback (before or after) execution of the instruction
        :type pos: CbPos
        :param opcode: Opcode where to trigger the callback
        :type opcode: :py:obj:`bytes`
        :param callback: callback function
        :type callback: :py:obj:`tritondse.callbacks.OpcodeCallback`
        """
        if opcode not in self._opcode_cbs:
            self._opcode_cbs[opcode] = {CbPos.BEFORE: [], CbPos.AFTER: []}

        self._opcode_cbs[opcode][pos].append(callback)
        self._empty = False


    def register_pre_opcode_callback(self, opcode: bytes, callback: OpcodeCallback) -> None:
        """
        Register pre-opcode callback.

        :param opcode: Opcode where to trigger the callback
        :type opcode: :py:obj:`bytes`
        :param callback: callback function
        :type callback: :py:obj:`tritondse.callbacks.OpcodeCallback`
        """
        self.register_opcode_callback(CbPos.BEFORE, opcode, callback)


    def register_post_opcode_callback(self, opcode: bytes, callback: OpcodeCallback) -> None:
        """
        Register post-opcode callback.

        :param opcode: Opcode where to trigger the callback
        :type opcode: :py:obj:`bytes`
        :param callback: callback function
        :type callback: :py:obj:`tritondse.callbacks.OpcodeCallback`
        """
        self.register_opcode_callback(CbPos.AFTER, opcode, callback)


    def get_opcode_callbacks(self, opcode: bytes) -> Tuple[List[OpcodeCallback], List[OpcodeCallback]]:
        """
        Get all the pre/post callbacks for a given opcode.

        :param opcode: Opcode where to trigger the callback
        :type opcode: :py:obj:`bytes`
        :return: tuple of lists containing callback functions for pre/post respectively
        """
        cbs = self._opcode_cbs.get(opcode, None)
        if cbs:
            return cbs[CbPos.BEFORE], cbs[CbPos.AFTER]
        else:
            return [], []


    def register_mnemonic_callback(self, pos: CbPos, mnemonic: OPCODE, callback: MnemonicCallback) -> None:
        """
        Register a callback function on a given mnemonic before or after the execution
        of the associated instruction.

        :param pos: When to trigger the callback (before or after) execution of the instruction
        :type pos: CbPos
        :param mnemonic: Mnemonic where to trigger the callback
        :type mnemonic: :py:obj:`OPCODE`
        :param callback: callback function
        :type callback: :py:obj:`tritondse.callbacks.MnemonicCallback`
        """
        if mnemonic not in self._mnemonic_cbs:
            self._mnemonic_cbs[mnemonic] = {CbPos.BEFORE: [], CbPos.AFTER: []}

        self._mnemonic_cbs[mnemonic][pos].append(callback)
        self._empty = False


    def register_pre_mnemonic_callback(self, mnemonic: OPCODE, callback: MnemonicCallback) -> None:
        """
        Register pre-mnemonic callback.

        :param mnemonic: Mnemonic where to trigger the callback
        :type mnemonic: :py:obj:`OPCODE`
        :param callback: callback function
        :type callback: :py:obj:`tritondse.callbacks.MnemonicCallback`
        """
        self.register_mnemonic_callback(CbPos.BEFORE, mnemonic, callback)


    def register_post_mnemonic_callback(self, mnemonic: OPCODE, callback: MnemonicCallback) -> None:
        """
        Register post-mnemonic callback.

        :param mnemonic: Mnemonic where to trigger the callback
        :type mnemonic: :py:obj:`OPCODE`
        :param callback: callback function
        :type callback: :py:obj:`tritondse.callbacks.MnemonicCallback`
        """
        self.register_mnemonic_callback(CbPos.AFTER, mnemonic, callback)


    def get_mnemonic_callbacks(self, mnemonic: OPCODE) -> Tuple[List[MnemonicCallback], List[MnemonicCallback]]:
        """
        Get all the pre/post callbacks for a given mnemonic.

        :param mnemonic: Mnemonic where to trigger the callback
        :type mnemonic: :py:obj:`OPCODE`
        :return: tuple of lists containing callback functions for pre/post respectively
        """
        cbs = self._mnemonic_cbs.get(mnemonic, None)
        if cbs:
            return cbs[CbPos.BEFORE], cbs[CbPos.AFTER]
        else:
            return [], []


    def register_function_callback(self, func_name: str, callback: AddrCallback) -> None:
        """
        Register a callback on the address of the given function name.
        The function name is only resolve when the callback manager is binded
        to a SymbolicExecutor.

        :param func_name: Function name
        :type func_name: str
        :param callback: callback function
        :type callback: :py:obj:`tritondse.callbacks.AddrCallback`
        """
        if func_name in self._func_to_register:
            self._func_to_register[func_name].append(callback)
        else:
            self._func_to_register[func_name] = [callback]


    def register_instruction_callback(self, pos: CbPos, callback: InstrCallback) -> None:
        """
        Register a callback triggered on each instruction executed, before or after its
        side effects have been applied to ProcessState.

        :param pos: before, after execution of the instruction
        :type pos: CbPos
        :param callback: callback function to trigger
        :type callback: :py:obj:`tritondse.callbacks.InstrCallback`
        """
        self._instr_cbs[pos].append(callback)
        self._empty = False


    def register_pre_instruction_callback(self, callback: InstrCallback) -> None:
        """
        Register a pre-execution callback on all instruction executed by the engine.

        :param callback: callback function to trigger
        :type callback: :py:obj:`tritondse.callbacks.InstrCallback`
        """
        self.register_instruction_callback(CbPos.BEFORE, callback)


    def register_post_instruction_callback(self, callback: InstrCallback) -> None:
        """
        Register a post-execution callback on all instruction executed by the engine.

        :param callback: callback function to trigger
        :type callback: :py:obj:`tritondse.callbacks.InstrCallback`
        """
        self.register_instruction_callback(CbPos.AFTER, callback)


    def get_instruction_callbacks(self) -> Tuple[List[InstrCallback], List[InstrCallback]]:
        """
        Get all the pre/post callbacks for instructions.

        :return: tuple of lists containing callback functions for pre/post respectively
        """
        return self._instr_cbs[CbPos.BEFORE], self._instr_cbs[CbPos.AFTER]


    def register_pre_execution_callback(self, callback: SymExCallback) -> None:
        """
        Register a callback executed after program loading, registers and memory
        initialization. Thus this callback is called just before executing the
        first instruction.

        :param callback: Callback function to trigger
        :type callback: :py:obj:`tritondse.callbacks.SymExCallback`
        """
        self._pre_exec.append(callback)
        self._empty = False


    def register_post_execution_callback(self, callback: SymExCallback) -> None:
        """
        Register a callback executed after program loading, registers and memory
        initialization. Thus this callback is called after executing upon program
        exit (or crash)

        :param callback: Callback function to trigger
        :type callback: :py:obj:`tritondse.callbacks.SymExCallback`
        """
        self._post_exec.append(callback)
        self._empty = False


    def register_exploration_step_callback(self, callback: ExplorationStepCallback) -> None:
        """
        Register a callback executed before each exploration step. The object
        given in parameter is the SymbolicExplorator itself.

        :param callback: Callback function to trigger
        :type callback: :py:obj:`tritondse.callbacks.ExplorationStepCallback`
        """
        self._step_cbs.append(callback)
        # self._empty = False  # Does not impact the emptiness of the callbackmanager

    def get_execution_callbacks(self) -> Tuple[List[SymExCallback], List[SymExCallback]]:
        """
        Get all the pre/post callbacks for the current symbolic execution.

        :return: tuple of lists containing callback functions for pre/post respectively
        """
        return self._pre_exec, self._post_exec


    def register_memory_read_callback(self, callback: MemReadCallback) -> None:
        """
        Register a callback that will be triggered by any read in the concrete
        memory of the process state.

        :param callback: Callback function to be called
        :type callback: :py:obj:`tritondse.callbacks.MemCallback`
        """
        self._mem_read_cbs.append(callback)
        self._empty = False


    def register_memory_write_callback(self, callback: MemWriteCallback) -> None:
        """
        Register a callback called on each write in the concrete memory state
        of the process.

        :param callback: Callback function to be called
        :type callback: :py:obj:`tritondse.callbacks.MemCallback`
        """
        self._mem_write_cbs.append(callback)
        self._empty = False


    def register_register_read_callback(self, callback: RegReadCallback) -> None:
        """
        Register a callback on each register read during the symbolic execution.

        :param callback: Callback function to be called
        :type callback: :py:obj:`tritondse.callbacks.RegReadCallback`
        """
        self._reg_read_cbs.append(callback)
        self._empty = False


    def register_register_write_callback(self, callback: RegWriteCallback) -> None:
        """
        Register a callback on each register write during the symbolic execution.

        :param callback: Callback function to be called
        :type callback: :py:obj:`tritondse.callbacks.RegReadCallback`
        """
        self._reg_write_cbs.append(callback)
        self._empty = False


    def register_thread_context_switch_callback(self, callback: ThreadCallback) -> None:
        """
        Register a callback triggered upon each thread context switch during the execution.

        :param callback: Callback to be called
        :type callback: :py:obj:`tritondse.callbacks.ThreadCallback`
        """
        self._ctx_switch.append(callback)
        self._empty = False


    def get_context_switch_callback(self) -> List[ThreadCallback]:
        """
        Get the list of all function callback to call when thread is being scheduled.

        :return: List of callbacks defined when thread is being scheduled
        """
        return self._ctx_switch


    def register_new_input_callback(self, callback: NewInputCallback) -> None:
        """
        Register a callback function called when the SMT solver find a new model namely
        a new input. This callback is called before any treatment on the input (worklist, etc.).
        It thus allow to post-process the input before it getting put in the queue.

        :param callback: callback function
        :type callback: :py:obj:`tritondse.callbacks.NewInputCallback`
        """
        self._new_input_cbs.append(callback)
        self._empty = False


    def get_new_input_callback(self) -> List[NewInputCallback]:
        """
        Get the list of all function callback to call when an a new
        input is generated by SMT.

        :return: List of callbacks to call on input generation
        """
        return self._new_input_cbs

    def register_on_solving_callback(self, callback: SymExSolvingCallback) -> None:
        """
        Register a callback function called when a branch is about to
        be solved. This callback is called before the branch is solved and will
        use the result of the callback to go ahead with the solving or skip it.

        :param callback: callback function
        :type callback: :py:obj:`tritondse.callbacks.BranchSolvingCallback`
        """
        self._branch_solving_cbs.append(callback)
        self._empty = False

    def get_on_solving_callback(self) -> List[SymExSolvingCallback]:
        """
        Get the list of all function callbacks to call when a branch is about
        to be solved.

        :return: List of callbacks to call on branch solving
        """
        return self._branch_solving_cbs

    def register_on_branch_covered_callback(self, callback: BranchCoveredCallback) -> None:
        """
        Register a callback function called when a branch covered. This callback
        is called after the branch is solved.

        :param callback: callback function
        :type callback: :py:obj:`tritondse.callbacks.BranchCoveredCallback`
        """
        self._branch_covered_cbs.append(callback)
        self._empty = False


    def get_on_branch_covered_callback(self) -> List[BranchCoveredCallback]:
        """
        Get the list of all function callbacks to call when a branch is about
        to be solved.

        :return: List of callbacks to call on branch covered
        """
        return self._branch_covered_cbs


    def register_memory_violation_callback(self, callback: MemoryViolationCallback) -> None:
        """
        Register a callback function called when a memory violation occured during the
        emulation.

        :param callback: callback function
        :type callback: :py:obj:`tritondse.callbacks.MemoryViolationCallback`
        """
        self._mem_violation_cbs.append(callback)
        self._empty = False

    def get_memory_violation_callbacks(self) -> List[MemoryViolationCallback]:
        """
        Get all memory violation callbacks.

        :return: list of memory violation callbacks
        """
        return self._mem_violation_cbs


    def get_exploration_step_callbacks(self) -> List[ExplorationStepCallback]:
        """
        Get all the exploration step callbacks

        :return: list containing callbacks
        """
        return self._step_cbs

    def register_pre_imported_routine_callback(self, routine_name: str, callback: RtnCallback) -> None:
        """
        Register a callback before call to an imported routines

        :param routine_name: the routine name
        :param callback: callback function
        :type callback: :py:obj:`tritondse.callbacks.RtnCallback`
        """
        if routine_name in self._pre_rtn_cbs:
            self._pre_rtn_cbs[routine_name].append(callback)
        else:
            self._pre_rtn_cbs[routine_name] = [callback]
        self._empty = False


    def register_post_imported_routine_callback(self, routine_name: str, callback: RtnCallback) -> None:
        """
        Register a callback, called after the call to imported routines.

        :param routine_name: the routine name
        :param callback: callback function
        :type callback: :py:obj:`tritondse.callbacks.RtnCallback`
        """
        if routine_name in self._post_rtn_cbs:
            self._post_rtn_cbs[routine_name].append(callback)
        else:
            self._post_rtn_cbs[routine_name] = [callback]
        self._empty = False


    def get_imported_routine_callbacks(self, routine_name: str) -> Tuple[List[RtnCallback], List[RtnCallback]]:
        """
        Get the list of all callbacks for an imported routine

        :param routine_name: the routine name
        :return: Tuple of list of callbacks (for pre and post)
        """
        pre_ret = (self._pre_rtn_cbs[routine_name] if routine_name in self._pre_rtn_cbs else [])
        post_ret = (self._post_rtn_cbs[routine_name] if routine_name in self._post_rtn_cbs else [])
        return pre_ret, post_ret


    def register_probe(self, probe: ProbeInterface) -> None:
        """
        Register a probe. That function will iterate the ``cbs`` attribute
        of the object, and will register each entries in self.

        .. warning:: Does not implement all CbType

        :param probe: a probe interface
        :type probe: ProbeInterface
        """
        for kind, cb, arg in probe.callbacks:
            try:
                mapping_with_args = {
                    CbType.PRE_RTN: self.register_pre_imported_routine_callback,
                    CbType.POST_RTN: self.register_post_imported_routine_callback,
                    CbType.PRE_ADDR: self.register_pre_addr_callback,
                    CbType.POST_ADDR: self.register_post_addr_callback,
                    CbType.PRE_MNEM: self.register_pre_mnemonic_callback,
                    CbType.POST_MNEM: self.register_post_mnemonic_callback,
                    CbType.PRE_OPCODE: self.register_pre_opcode_callback,
                    CbType.POST_OPCODE: self.register_post_opcode_callback,
                }
                mapping_with_args[kind](arg, cb)
            except KeyError:
                mapping = {
                    CbType.CTX_SWITCH: self.register_thread_context_switch_callback,
                    CbType.MEMORY_READ: self.register_memory_read_callback,
                    CbType.MEMORY_WRITE: self.register_memory_write_callback,
                    CbType.POST_EXEC: self.register_post_execution_callback,
                    CbType.POST_INST: self.register_post_instruction_callback,
                    CbType.PRE_EXEC: self.register_pre_execution_callback,
                    CbType.PRE_INST: self.register_pre_instruction_callback,
                    CbType.REG_READ: self.register_register_read_callback,
                    CbType.REG_WRITE: self.register_register_write_callback,
                    CbType.NEW_INPUT: self.register_new_input_callback,
                    CbType.EXPLORE_STEP: self.register_exploration_step_callback,
                    CbType.BRANCH_COV: self.register_on_branch_covered_callback,
                    CbType.SYMEX_SOLVING: self.register_on_solving_callback,
                    CbType.MEM_VIOLATION: self.register_memory_violation_callback
                }
                mapping[kind](cb)


    def fork(self) -> 'CallbackManager':
        """
        Fork the current CallbackManager in a new object instance
        (that will be unbinded). That method is used by the SymbolicExplorator
        to ensure each SymbolicExecutor running concurrently will have
        their own instance off the CallbackManager.

        :return: Fresh instance of CallbackManager
        :rtype: CallbackManager
        """
        cbs = CallbackManager()

        # SymbolicExecutor callbacks
        cbs._pc_addr_cbs        = self._pc_addr_cbs.copy()
        cbs._opcode_cbs         = self._opcode_cbs.copy()
        cbs._mnemonic_cbs       = self._mnemonic_cbs.copy()
        cbs._instr_cbs          = self._instr_cbs.copy()
        cbs._pre_exec           = self._pre_exec.copy()
        cbs._post_exec          = self._post_exec.copy()
        cbs._ctx_switch         = self._ctx_switch.copy()
        cbs._new_input_cbs      = self._new_input_cbs.copy()
        cbs._branch_solving_cbs = self._branch_solving_cbs.copy()
        cbs._branch_covered_cbs = self._branch_covered_cbs.copy()
        cbs._pre_rtn_cbs        = self._pre_rtn_cbs.copy()
        cbs._post_rtn_cbs       = self._post_rtn_cbs.copy()
        cbs._mem_violation_cbs  = self._mem_violation_cbs.copy()
        # Triton callbacks
        cbs._mem_read_cbs  = self._mem_read_cbs.copy()
        cbs._mem_write_cbs = self._mem_write_cbs.copy()
        cbs._reg_read_cbs  = self._reg_read_cbs.copy()
        cbs._reg_write_cbs = self._reg_write_cbs.copy()
        cbs._empty         = self._empty
        # Copy temporary data
        cbs._func_to_register = self._func_to_register.copy()

        return cbs

    def unregister_callback(self, callback: Callable) -> None:
        """
        Unregister the given callback from the manager.

        :param callback: callback to remove
        :return: None
        """
        for addr, itms in self._pc_addr_cbs.items():
            for loc in CbPos:
                if callback in itms[loc]:
                    itms[loc].remove(callback)

        for opcode, itms in self._opcode_cbs.items():
            for loc in CbPos:
                if callback in itms[loc]:
                    itms[loc].remove(callback)

        for mnemonic, itms in self._mnemonic_cbs.items():
            for loc in CbPos:
                if callback in itms[loc]:
                    itms[loc].remove(callback)

        for loc in CbPos:
            if callback in self._instr_cbs[loc]:
                self._instr_cbs[loc].remove(callback)

        for cb_list in [self._step_cbs, self._pre_exec, self._post_exec, self._ctx_switch, self._new_input_cbs, self._branch_solving_cbs, self._branch_covered_cbs,
                        self._mem_read_cbs, self._mem_write_cbs, self._reg_read_cbs, self._reg_write_cbs, self._mem_violation_cbs]:
            if callback in cb_list:
                cb_list.remove(callback)

        for d in [self._pre_rtn_cbs, self._post_rtn_cbs]:
            for cb_list in d.values():
                if callback in cb_list:
                    cb_list.remove(callback)

    def reset(self) -> None:
        """
        Reset all callbacks
        :return:
        """
        # SymbolicExplorator callbacks
        self._step_cbs = []  # Callback called between each exploration steps

        # SymbolicExecutor callbacks
        self._pc_addr_cbs        = {}  # addresses reached
        self._opcode_cbs         = {}  # opcode before and after
        self._mnemonic_cbs       = {}  # mnemonic before and after
        self._instr_cbs          = {CbPos.BEFORE: [], CbPos.AFTER: []}  # all instructions
        self._pre_exec           = []  # before execution
        self._post_exec          = []  # after execution
        self._ctx_switch         = []  # on each thread context switch (implementing pre/post?)
        self._new_input_cbs      = []  # each time an SMT model is get
        self._branch_solving_cbs = []  # each time a covitem is about to be solved
        self._branch_covered_cbs = []  # each time a covitem is covered
        self._pre_rtn_cbs        = {}  # before imported routine calls ({str: [RtnCallback]})
        self._post_rtn_cbs       = {}  # after imported routine calls ({str: [RtnCallback]})
        self._mem_violation_cbs  = []  # reset the memory violation calls

        # Triton callbacks
        self._mem_read_cbs  = []  # memory reads
        self._mem_write_cbs = []  # memory writes
        self._reg_read_cbs  = []  # register reads
        self._reg_write_cbs = []  # register writes
        self._empty         = True



# Built-in imports
from pathlib import Path
from typing import Union, Generator, Tuple, Optional, Any, List

# third-party imports
import quokka
import networkx
import lief

# local imports
from tritondse.loaders import Program, LoadableSegment
from tritondse.coverage import CoverageSingleRun
from tritondse.types import PathLike, Addr, Architecture, Platform, Endian


class QuokkaProgram(quokka.Program):
    def __init__(self, export_file: Union[Path, str], exec_path: Union[Path, str]):
        super(QuokkaProgram, self).__init__(export_file, exec_path)

        self.program = Program(self.executable.exec_file.as_posix())

    def get_call_graph(self, backedge_on_ret=False) -> networkx.DiGraph:
        """
        Compute the call graph of the program.

        :param backedge_on_ret: if true, add a back edge to represent the "return"
        :return: call graph as a digraph
        """
        g = networkx.DiGraph()
        for fun in self.values():
            g.add_edges_from((fun.start, x.start) for x in fun.calls)
            if backedge_on_ret:  # Add return edge
                g.add_edges_from((x.start, fun.start) for x in fun.calls)
        return g

    @staticmethod
    def get_slice(graph: networkx.DiGraph, frm: Any, to: Any) -> networkx.DiGraph:
        """
        Compute the slice between the two nodes on the given graph.
        The slice is the intersection of reachable node (of ``frm``) and
        ancestors of ``to``. The result is a subgraph of the original graph.

        :param graph: Graph on which to compute the slice
        :param frm: node identifier
        :param to: node identifier
        :return: sub graph
        """
        succs = networkx.descendants(graph, frm)
        preds = networkx.ancestors(graph, to)
        return graph.subgraph(succs.intersection(preds).union({frm, to}))

    def merge(self, coverage: CoverageSingleRun):
        # TODO: To implement
        raise NotImplementedError

    def __repr__(self):
        return f"<{self.export_file.name}  funs:{len(self)}>"

    def get_caller_instructions(self, target: quokka.Function) -> List[int]:
        """Get the list of instructions calling `target`
        """

        # Get the first instruction of the target
        first_inst = target.get_instruction(target.start)
        assert first_inst is not None

        # Reference holder
        ref = target.program.references

        caller_instructions = []
        for reference in ref.resolve_inst_instance(first_inst.inst_tuple, quokka.types.ReferenceType.CALL, towards=True):
            _, block, offset = reference.source
            inst = list(block.instructions)[offset]
            caller_instructions.append(inst.address)
        return caller_instructions


    # ============== Methods for interroperability with Program object ==============
    @property
    def path(self) -> Path:
        return self.program.path

    @path.setter
    def path(self, path: Path) -> None:
        self.program.path = path

    @property
    def entry_point(self) -> Addr:
        return self.program.entry_point

    @property
    def architecture(self) -> Architecture:
        return self.program.architecture

    @property
    def platform(self) -> Platform:
        return self.program.platform

    @property
    def endianness(self) -> Endian:
        return self.program.endianness

    @property
    def format(self) -> lief.EXE_FORMATS:
        return self.program.format

    @property
    def relocation_enum(self):
        return self.program.relocation_enum

    def memory_segments(self) -> Generator[LoadableSegment, None, None]:
        return self.program.memory_segments()

    def imported_functions_relocations(self) -> Generator[Tuple[str, Addr], None, None]:
        return self.program.imported_functions_relocations()

    def imported_variable_symbols_relocations(self) -> Generator[Tuple[str, Addr], None, None]:
        return self.program.imported_variable_symbols_relocations()

    def find_function_addr(self, name: str) -> Optional[Addr]:
        return self.program.find_function_addr(name)

    def find_function_from_addr(self, address: Addr) -> Optional[quokka.function.Function]:
        for f in self.values():
            if f.in_func(address):
                return f



from __future__ import annotations

import sys
from enum import IntEnum, Enum, auto, IntFlag
from pathlib import Path
from triton import ARCH, SOLVER_STATE, SOLVER
from typing import Union, TypeVar, Tuple, Type
import io
import enum_tools.documentation
from dataclasses import dataclass

PathLike = Union[str, Path]
"""Type representing file either as a file, either as a Path object"""

Addr = int
"""Integer representing an address"""

rAddr = int
"""Integer representing a relative address"""

BitSize = int
"""Integer representing a value in bits"""

ByteSize = int
"""Integer representing a value in bytes"""

Input = bytes
""" Type representing an Input (which is bytes) """

Register = TypeVar('Register')
"""Register identifier as used by Triton *(not the Register object itself)*"""

Registers = TypeVar('Registers')
"""Set of registers as used by Triton"""

PathConstraint = TypeVar('PathConstraint')
""" `PathConstraint <https://triton.quarkslab.com/documentation/doxygen/py_PathConstraint_page.html>`_ object as returned by Triton"""

AstNode = TypeVar('AstNode')
""" SMT logic formula as returned by Triton (`AstNode <https://triton.quarkslab.com/documentation/doxygen/py_AstNode_page.html>`_) """

Model = TypeVar('Model')
""" Solver `Model <https://triton.quarkslab.com/documentation/doxygen/py_SolverModel_page.html>`_ as returned by Triton """

Expression = TypeVar('Expression')
""" Symbolic Expression as returned by Triton (`SymbolicExpression <https://triton.quarkslab.com/documentation/doxygen/py_SymbolicExpression_page.html>`_) """

SymbolicVariable = TypeVar('SymbolicVariable')
""" Symbolic Variable as returned by Triton (`SymbolicExpression <https://triton.quarkslab.com/documentation/doxygen/py_SymbolicVariable_page.html>`_) """

Edge = Tuple[Addr, Addr]
""" Type representing a edge in the program """

PathHash = str
"""Type representing the hash of path to uniquely identify any path """


@enum_tools.documentation.document_enum
class SymExType(str, Enum):
    """
    Symobolic Expression type enum. (internal usage only)
    """

    CONDITIONAL_JMP = 'cond-jcc'  # doc: symbolic expression is a conditional jump
    DYNAMIC_JMP = 'dyn-jmp'       # doc: symbolic expression is a dynamic jump
    SYMBOLIC_READ = 'sym-read'    # doc: symbolic expression is a symbolic memory read
    SYMBOLIC_WRITE = 'sym-write'  # doc: symbolic expression is a symbolic memory write


if sys.version_info.minor >= 8:
    from typing import TypedDict

    class PathBranch(TypedDict):
        """
        Typed dictionnary describing the branch information
        returned by Triton (with getBranchConstraints())
        """
        isTaken: bool
        srcAddr: Addr
        dstAddr: Addr
        constraint: AstNode
else:
    PathBranch = TypeVar('PathBranch')
    """ PathBranchobject as returned by Triton.
    Thus it is a dictionnary with the keys:
    """


@enum_tools.documentation.document_enum
class Architecture(IntEnum):
    """
    Common architecture Enum fully compatible with Triton
    `ARCH <https://triton.quarkslab.com/documentation/doxygen/py_ARCH_page.html>`_
    """
    AARCH64 = ARCH.AARCH64  # doc: Aarch64 architecture
    ARM32 = ARCH.ARM32      # doc: ARM architecture (32 bits)
    X86 = ARCH.X86          # doc: x86 architecture (32 bits)
    X86_64 = ARCH.X86_64    # doc: x86-64 architecture (64 bits)

@enum_tools.documentation.document_enum
class ArchMode(IntFlag):
    """
    Various architecture specific modes that can be enabled or disabledd.
    (meant to be fullfilled)
    """
    THUMB = 1   # doc: set thumb mode for ARM32 architecture


@enum_tools.documentation.document_enum
class Platform(IntEnum):
    """
    Platform associated to a binary
    """
    LINUX = auto()    # doc: Linux platform
    WINDOWS = auto()  # doc: Windows platform
    MACOS = auto()    # doc: Mac OS platform
    ANDROID = auto()  # doc: Android platform
    IOS = auto()      # doc: IOS platform

@enum_tools.documentation.document_enum
class SmtSolver(IntEnum):
    """ Common SMT Solver Enum fully compatible with Triton """
    Z3 = SOLVER.Z3              # doc: Z3 SMT solver
    BITWUZLA = SOLVER.BITWUZLA  # doc: bitwuzla solver


@enum_tools.documentation.document_enum
class SolverStatus(IntEnum):
    """ Common Solver Enum fully compatible with Triton ARCH """
    SAT     = SOLVER_STATE.SAT      # doc: Formula is satisfiable (SAT)
    UNSAT   = SOLVER_STATE.UNSAT    # doc: Formula is unsatisfiable (UNSAT)
    TIMEOUT = SOLVER_STATE.TIMEOUT  # doc: Formula solving did timeout
    UNKNOWN = SOLVER_STATE.UNKNOWN  # doc: Formula solving failed


@enum_tools.documentation.document_enum
class Perm(IntFlag):
    """
    Flags encoding permissions. Used for memory pages.
    They can be combined as flags. e.g:

    .. code-block:: python

        rw = Perm.R | Perm.W
    """
    R = 4  # doc: Read
    W = 2  # doc: Write
    X = 1  # doc: Execute


@enum_tools.documentation.document_enum
class Endian(IntEnum):
    """
    Endianess of the binary.
    """
    LITTLE = 1  # doc: Little-endian
    BIG = 2     # doc: Big-endian


@dataclass
class FileDesc:
    """
    Type representing a file descriptor
    """
    """ The target program's file descriptor """
    id: int
    """ Name of the file """
    name: str
    """ The python file stream object """
    fd: io.IOBase

    @property
    def offset(self) -> int:
        return self.fd.tell()

    def is_real_fd(self) -> bool:
        return isinstance(self.fd, io.TextIOWrapper)

    def is_input_fd(self) -> bool:
        return isinstance(self.fd, io.BytesIO)

    def fgets(self, max_size: int) -> bytes:
        s = b""
        for i in range(max_size):
            c = self.fd.read(1)
            if not c:  # EOF
                break
            c = c if isinstance(c, bytes) else c.encode()
            s += c
            if c == b"\x00":
                return s
            elif c == b"\n":
                break
        # If get there read max_size
        return s+b"\x00"

    def read(self, size: int) -> bytes:
        data = self.fd.read(size)
        return data if isinstance(data, bytes) else data.encode()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Installation of the tritondse module."""

import sys
from setuptools import setup, find_packages


with open("README.md") as f:
    README = f.read()


setup(
    name="tritondse",
    version="0.1.9",
    description="A library of Dynamic Symbolic Exploration based the Triton library",
    packages=find_packages(),
    long_description=README,
    long_description_content_type='text/markdown',
    url="https://github.com/quarkslab/tritondse",
    project_urls={
        "Documentation": "https://quarkslab.github.io/tritondse/",
        "Bug Tracker": "https://github.com/quarkslab/tritondse/issues",
        "Source": "https://github.com/quarkslab/tritondse"
    },
    setup_requires=[],
    install_requires=[
        "triton-library",
        "lief",
        "pyQBDI",
        "cle",
        "quokka-project",
        "enum_tools"
    ],
    tests_require=[],
    license="AGPL-3.0",
    author="Quarkslab",
    classifiers=[
        'Topic :: Security',
        'Environment :: Console',
        'Operating System :: OS Independent',
    ],
    test_suite="",
    scripts=[]
)



# built-in imports
import json
from typing import Optional

# Local imports
from tritondse.seed import Seed
from tritondse.coverage import GlobalCoverage
from tritondse.workspace import Workspace
import tritondse.logging

logger = tritondse.logging.get("seedmanager")


class SeedScheduler:
    """
    Abstract class for all seed selection strategies.
    This class provides the base methods that all
    subclasses should implement to be compliant with
    the interface.
    """

    def has_seed_remaining(self) -> bool:
        """
        Returns true if there are still seeds to be processed in the scheduler

        :returns: true if there are seeds to process
        """
        raise NotImplementedError()

    def add(self, seed: Seed) -> None:
        """
        Add a new seed in the scheduler

        :param seed: Seed to add in the scheduler
        :type seed: Seed
        """
        raise NotImplementedError()

    def update_worklist(self, coverage: GlobalCoverage) -> None:
        """
        Call after every execution.
        That function might help the scheduler with some of its internal states.
        For instance the scheduler is keep somes seed meant to cover an address
        which is now covered, it can just drop these seeds.

        :param coverage: global coverage of the exploration
        :type coverage: GlobalCoverage
        """
        raise NotImplementedError()

    def can_solve_models(self) -> bool:
        """
        Function called by the seed manager to know if it can
        start negating branches to discover new paths. Some seed
        scheduler might want to run concretely all inputs first
        before starting negating branches.

        :return: true if the :py:obj:`SeedManager` can negate branches
        """
        raise NotImplementedError()

    def pick(self) -> Optional[Seed]:
        """
        Return the next seed to execute.

        :returns: seed to execute
        :rtype: Seed
        """
        raise NotImplementedError()

    def post_execution(self) -> None:
        """
        Called at the end of each execution after the generation of new seeds through SMT.
        Last thing called before starting the next iteration.
        """
        pass

    def post_exploration(self, workspace: Workspace) -> None:
        """
        Called at the end of the exploration to perform
        some clean-up or anything else.
        """
        pass


class WorklistAddressToSet(SeedScheduler):
    """
    This worklist classifies seeds by addresses. We map a seed X to an
    address Y, if the seed X has been generated to reach the address Y.
    When the method pick() is called, seeds covering a new address 'Y'
    are selected first. Otherwise anyone is taken.
    """
    def __init__(self, manager: 'SeedManager'):
        self.manager = manager
        self.cov = None
        self.worklist = dict() # {CovItem: set(Seed)}

    def __len__(self) -> int:
        """ Number of pending seeds to execute """
        count = 0
        for k in list(self.worklist):  # Copy in list to avoid race-condition
            v = self.worklist[k]
            count += len(v)
        return count

    def has_seed_remaining(self) -> bool:
        """ Returns true if there are still seeds in the worklist """
        return len(self) != 0

    def add(self, seed: Seed) -> None:
        """ Add a seed to the worklist """
        for obj in seed.coverage_objectives:
            if obj in self.worklist:
                self.worklist[obj].add(seed)
            else:
                self.worklist[obj] = {seed}

    def update_worklist(self, coverage: GlobalCoverage) -> None:
        """ Update the coverage state of the woklist with the global one """
        self.cov = coverage

    def can_solve_models(self) -> bool:
        """
        Always true.
        This strategy always allows solving branches. As a consequence
        it might try to solve a branch already covered in a seed not run yet.
        But this enables iterating a seed only once.

        :returns: True
        """
        return True

    def pick(self) -> Optional[Seed]:
        """ Return the next seed to execute

        :returns: next seed to execute (first one covering new addresses, otherwise any other)
        :rtype: Seed
        """
        seed_picked = None
        item_picked = None
        to_remove = set()

        for k, v in self.worklist.items():
            # If the set is empty remove the entry
            if not len(v):
                to_remove.add(k)
                continue

            # If the address has never been executed, return the seed
            if not self.cov.is_covered(k):
                seed_picked = v.pop()
                item_picked = k
                if not len(v):
                    to_remove.add(k)
                break

        # If all adresses has been executed, just pick a random seed
        if not seed_picked:
            for k, v in self.worklist.items():
                if v:
                    seed_picked = v.pop()
                    item_picked = k
                    if not len(v):
                        to_remove.add(k)
                    break

        # If we did not have a seed again the worklist is definitely empty
        if not seed_picked:
            return None

        # Pop the seed from all worklist[X] where it is
        for obj in seed_picked.coverage_objectives:
            if obj != item_picked:   # already poped it from item_picked thus only pop the other
                if obj in self.worklist and seed_picked in self.worklist[obj]: 
                    self.worklist[obj].remove(seed_picked)
                    if not self.worklist[obj]:
                        to_remove.add(obj)

        # Garbage the worklist
        for i in to_remove:
            self.worklist.pop(i)

        return seed_picked


class WorklistRand(SeedScheduler):
    """
    Trivial strategy that returns any Seed without any classification.
    It uses a Set for insertion and pop (which is random) for picking seeds.
    """
    def __init__(self, manager: 'SeedManager'):
        self.worklist = set() # set(Seed)

    def __len__(self) -> int:
        """ Number of pending seeds to execute """
        return len(self.worklist)

    def has_seed_remaining(self) -> bool:
        """ Returns true if there are still seeds in the worklist """
        return len(self) != 0

    def add(self, seed: Seed) -> None:
        """ Add a seed to the worklist

        :param seed: Seed to add to this rand scheduler
        :type seed: Seed
        """
        self.worklist.add(seed)

    def update_worklist(self, coverage: GlobalCoverage) -> None:
        """ Update the coverage state of the worklist with the global one """
        self.cov = coverage

    def can_solve_models(self) -> bool:
        """ Always true """
        return True

    def pick(self) -> Optional[Seed]:
        """
        Return the next seed to execute. The method pop() removes a random element
        from the set and returns the removed element. Unlike, a stack a
        random element is popped off the set.

        :returns: next seed to executre
        :rtype: Seed
        """
        return self.worklist.pop() if self.worklist else None


class FreshSeedPrioritizerWorklist(SeedScheduler):
    """
    Strategy that first execute all seeds without negating branches
    in order to get the most updated coverage and which then re-run
    all relevant seeds to negate their branches.

    This worklist works as follow:
        - return first fresh seeds first to get them executed (to improve coverage)
        - keep the seed in the worklist up until it gets dropped or thoroughtly processed
        - if no fresh seed is available, iterates seed that will generate coverage
    """
    def __init__(self, manager: 'SeedManager'):
        self.manager = manager
        self.fresh = []       # Seed never processed (list to make sure we can pop first one received)
        self.worklist = dict() # CovItem -> set(Seed)

    def __len__(self) -> int:
        """ Number of pending seeds to execute """
        s = set()
        for seeds in list(self.worklist.values()):
            s.update(seeds)
        return len(self.fresh) + len(s)

    def has_seed_remaining(self) -> bool:
        """ Returns true if there are still seeds in the worklist """
        return len(self) != 0

    def add(self, seed: Seed) -> None:
        """ Add a seed to the worklist

        :param seed: seed to add to the scheduler
        :type seed: Seed
        """
        if seed.coverage_objectives:  # If the seed already have coverage objectives
            for item in seed.coverage_objectives:  # Add it in our worklist
                if item in self.worklist:
                    self.worklist[item].add(seed)
                else:
                    self.worklist[item] = {seed}
            # seed.coverage_objectives.clear()  # Flush the objectives
        else:  # Otherwise it is fresh
            self.fresh.append(seed)

    def update_worklist(self, coverage: GlobalCoverage) -> None:
        """ Update the coverage state of the worklist with the global one """
        # Iterate the worklist to see if some items have now been covered
        # and are thus not interesting anymore
        to_remove = [x for x in self.worklist if coverage.is_covered(x)]

        for item in to_remove:
            for seed in self.worklist.pop(item):
                seed.coverage_objectives.remove(item)
                if not seed.coverage_objectives:  # The seed cannot improve the coverage of anything
                    self.manager.drop_seed(seed)

    def can_solve_models(self) -> bool:
        """
        Returns True if there are no "fresh" seeds to execute.

        :returns: True if all fresh seeds have been executed.
        """
        return not self.fresh

    def pick(self) -> Optional[Seed]:
        """ Return the next seed to execute """
        # Pop first fresh seed
        if self.fresh:
            return self.fresh.pop(0)  # Return first item as it is the older

        # Then pop seed meant to crash
        if ... in self.worklist:  # If we have specific seeds (mostly generated by sanitizers)
            it = self.worklist[...].pop()
            if not self.worklist[...]:  # Remove the key if now empty
                self.worklist.pop(...)
            return it

        if not self.worklist:
            return None

        # Then pop traditional coverage seeds
        k = list(self.worklist.keys())[0]      # arbitrary covitem
        seed = self.worklist[k].pop()          # remove first seed inside
        for it in seed.coverage_objectives:    # Remove the seed from all worklist[x]
            if it != k:                        # we already popped the item from k
                self.worklist[it].remove(seed) # remove the seed from that covitem set
            if not self.worklist[it]:          # remove the whole covitem if empty
                self.worklist.pop(it)
        return seed

    def post_execution(self) -> None:
        """
        Solely used to show intermediate statistics
        """
        logger.info(f"Seed Scheduler: worklist:{len(self)} Coverage objectives:{len(self.worklist)}  (fresh:{len(self.fresh)})")

    def post_exploration(self, workspace: Workspace) -> None:
        """
        At the end of the execution, print the worklist to know
        its state before exit.
        """
        workspace.save_metadata_file("coverage_objectives.json", json.dumps(list(self.worklist.keys())))



# built-in modules
import platform
from collections import namedtuple

# third-party module
from triton import OPCODE, TritonContext

# local imports
from tritondse.types import Architecture, Addr

Arch = namedtuple("Arch", "ret_reg pc_reg bp_reg sp_reg sys_reg reg_args halt_inst syscall_inst")

ARCHS = {
    Architecture.X86:     Arch('eax', 'eip', 'ebp', 'esp', 'eax',
                               [],
                               OPCODE.X86.HLT,
                               [OPCODE.X86.SYSCALL, OPCODE.X86.SYSENTER]),  # ignore int 80
    Architecture.X86_64:  Arch('rax', 'rip', 'rbp', 'rsp', 'rax',
                               ['rdi', 'rsi', 'rdx', 'rcx', 'r8', 'r9'],
                               OPCODE.X86.HLT,
                               [OPCODE.X86.SYSCALL, OPCODE.X86.SYSENTER]),  # ignore int 80
    Architecture.AARCH64: Arch('x0', 'pc', 'sp', 'sp', 'x8',
                               ['x0', 'x1', 'x2', 'x3', 'x4', 'x5', 'x6', 'x7'],
                               OPCODE.AARCH64.HLT,
                               [OPCODE.AARCH64.SVC]),
    Architecture.ARM32:   Arch('r0', 'pc', 'r11', 'sp', 'r7',
                               ['r0', 'r1', 'r2', 'r3'],
                               OPCODE.ARM32.HLT,
                               [OPCODE.ARM32.SVC])
}


class CpuState(dict):
    """
    Thin wrapper on a TritonContext, to allow accessing
    and modifying registers in a Pythonic way. It also
    abstract base, stack, and program counter for architecture
    agnostic operations. This class performs all actions
    on the TritonContext, and does not hold any information.
    It is just acting as a proxy

    .. note:: This class adds dynamically attributes corresponding
              to register. Thus attributes will vary from an architecture
              to the other.

    >>> cpu.rax
    12
    >>> cpu.rax += 1
    >>> cpu.rax
    13

    No data is stored, all operations are performed on the
    TritonContext:

    >>> cpu.__ctx.getConcreteRegisterValue(cpu.rsp)
    0x7ff6540
    >>> cpu.stack_pointer += 8
    >>> cpu.__ctx.getConcreteRegisterValue(cpu.rsp)
    0x7ff6548

    .. note:: The user is not meant to instanciate it manually, and must
              use it through :py:obj:`ProcessState`.
    """

    def __init__(self, ctx: TritonContext, arch_info: Arch):
        super(CpuState, self).__init__()
        self.__ctx = ctx
        self.__archinfo = arch_info
        for r in ctx.getAllRegisters():
            self[r.getName()] = r

    def __getattr__(self, name: str):
        """
        Return the concrete value of a given register name

        :param name: The name of the register
        :return: the concrete value of a given register
        """
        if name in self:
            return self.__ctx.getConcreteRegisterValue(self[name])
        else:
            super().__getattr__(name)

    def __setattr__(self, name: str, value: int):
        """
        Set a concrete value to a given register name

        :param name: The name of the register
        :param value: The concrete value to set
        """
        if name in self:
            self.__ctx.setConcreteRegisterValue(self[name], value)
        else:
            super().__setattr__(name, value)

    @property
    def program_counter(self) -> int:
        """
        :return: The value of the program counter (RIP for x86, PC for ARM ..)
        :rtype: int
        """
        return getattr(self, self.__archinfo.pc_reg)

    @program_counter.setter
    def program_counter(self, value: int) -> None:
        """
        Set a value to the program counter

        :param value: Value to set
        :type value: int
        """
        setattr(self, self.__archinfo.pc_reg, value)

    @property
    def base_pointer(self) -> int:
        """
        :return: The value of the base pointer register
        """
        return getattr(self, self.__archinfo.bp_reg)

    @base_pointer.setter
    def base_pointer(self, value: int) -> None:
        """
        Set a value to the base pointer register

        :param value: Value to set
        :return: None
        """
        setattr(self, self.__archinfo.bp_reg, value)

    @property
    def stack_pointer(self) -> int:
        """
        :return: The value of the stack pointer register
        """
        return getattr(self, self.__archinfo.sp_reg)

    @stack_pointer.setter
    def stack_pointer(self, value: int) -> None:
        """
        Set a value to the stack pointer register

        :param value: Value to set
        :type value: int
        """
        setattr(self, self.__archinfo.sp_reg, value)


def local_architecture() -> Architecture:
    """
    Returns the architecture of the local machine.

    :return: local architecture
    """
    arch_m = {"i386": Architecture.X86,
              "x86_64": Architecture.X86_64,
              "armv7l": Architecture.ARM32,
              "aarch64": Architecture.AARCH64}
    return arch_m[platform.machine()]



import time
import threading
import gc
from enum import Enum
from typing import Union, Type
from pathlib import Path
import stat
import enum_tools.documentation

from tritondse.config            import Config
from tritondse.process_state     import ProcessState
from tritondse.loaders           import Loader
from tritondse.seed              import Seed
from tritondse.seed              import SeedStatus
from tritondse.symbolic_executor import SymbolicExecutor
from tritondse.workspace         import Workspace
from tritondse.coverage          import GlobalCoverage
from tritondse.types             import Addr
from tritondse.exception         import StopExplorationException
from tritondse.seeds_manager import SeedManager
from tritondse.worklist import SeedScheduler
from tritondse.callbacks import CallbackManager
import tritondse.logging


logger = tritondse.logging.get("explorator")


@enum_tools.documentation.document_enum
class ExplorationStatus(Enum):
    """ Enum representing the current state of the exploration """
    NOT_RUNNING = 0  # doc: The explorator has not been started yet
    RUNNING     = 1  # doc: The explorator is running and performing exploration
    IDLE        = 2  # doc: The explorator is idle, it has run, but is out of inputs to process
    STOPPED     = 3  # doc: The explorator has been stopped by the user
    TERMINATED  = 4  # doc: The explorator has stopped normally, reached its goal (user-defined)


class SymbolicExplorator(object):
    """
    Symbolic Exploration. This class is in charge of iterating
    executions with the different seeds available in the workspace
    and generated along the way.
    """
    def __init__(self, config: Config, loader: Loader = None, workspace: Workspace = None, executor_stop_at: Addr = None, seed_scheduler_class: Type[SeedScheduler] = None):
        self.loader: Loader = loader  #: Program being analyzed
        self.config: Config = config  #: Configuration file
        self.cbm: CallbackManager = CallbackManager()  #: CallbackManager to register callbacks
        self._stop = False
        self.ts: int = time.time()  #: Timestamp (object instanciation)
        self.uid_counter: int = 0
        self.status: ExplorationStatus = ExplorationStatus.NOT_RUNNING  #: status of the execution
        self._executor_stop_at = executor_stop_at

        # Initialize the workspace
        if workspace:
            self.workspace: Workspace = workspace  #: exploration workspace
        else:
            self.workspace: Workspace = Workspace(self.config.workspace)  #: workspace object
            self.workspace.initialize(flush=False)

        # Save the configuration in the workspace
        self.workspace.save_file("config.json", self.config.to_json())

        # Save the binary in the workspace if not already done
        if self.loader:
            bin_path = self.workspace.get_binary_directory() / self.loader.bin_path.name
            if not bin_path.exists():  # If the program is not yet present
                self.workspace.save_file(bin_path, self.loader.bin_path.read_bytes())
                self.loader.bin_path = bin_path  # Patch its official new location
                bin_path.chmod(stat.S_IRWXU)  # Make it executable

        # Configure logfile in workspace
        tritondse.logging.enable_to_file(logger.level, self.workspace.logfile_path)

        # Initialize coverage
        self.coverage: GlobalCoverage = GlobalCoverage(self.config.coverage_strategy, self.config.branch_solving_strategy)
        """ GlobalCoverage object holding information about the global coverage.
        *(not really meant to be manipulated by the user)*
        """
        # Load workspace global coverage if any
        cov = self.workspace.get_metadata_file_path(GlobalCoverage.COVERAGE_FILE)
        if cov.exists():
            self.coverage = GlobalCoverage.from_file(cov)

        # Initialize the seed manager
        self.seeds_manager: SeedManager = SeedManager(self.coverage, self.workspace, self.config.smt_queries_limit, callback_manager=self.cbm, seed_scheduler_class=seed_scheduler_class)
        """ Manager of seed, holding all seeds related data and various statistics """

        # running executors (for debugging purposes)
        self.current_executor: SymbolicExecutor = None  #: last symbolic executor executed

        # General purpose attributes
        self._exec_count = 0
        self._total_emulation_time = 0

    @property
    def total_emulation_time(self) -> float:
        """ Represent total emulation time. This include all callbacks execution
        but not the SMT solving time (performed at the end). """
        return self._total_emulation_time

    @property
    def callback_manager(self) -> CallbackManager:
        """
        CallbackManager global instance that will be transmitted to
        all :py:obj:`SymbolicExecutor`.

        :rtype: CallbackManager
        """
        return self.cbm

    @property
    def execution_count(self) -> int:
        """
        Get the number of execution performed.

        :return: number of execution performed
        :rtype: int
        """
        return self._exec_count

    def __time_delta(self):
        return time.time() - self.ts

    def _worker(self, seed, uid):
        """ Worker thread """
        logger.info(f'Pick-up seed: {seed.filename} (fresh: {seed.is_fresh()})')

        if self.config.exploration_timeout and self.__time_delta() >= self.config.exploration_timeout:
            logger.info('Exploration timeout')
            self.stop_exploration()
            return

        # Execute the binary with seeds
        cbs = None if self.cbm.is_empty() else self.cbm.fork()
        logger.info(f"Initialize ProcessState with thread scheduling: {self.config.thread_scheduling}")
        execution = SymbolicExecutor(self.config, seed=seed, workspace=self.workspace, uid=uid, callbacks=cbs)
        if self.loader:  # If doing the exploration from a program
            execution.load(self.loader)
        else:
            execution.load_process(ProcessState())
        self.current_executor = execution

        # increment exec_count
        self._exec_count += 1

        ts = time.time()
        try:
            execution.run(self._executor_stop_at)
            expl_ts = time.time() - ts
        except StopExplorationException:
            expl_ts = time.time() - ts
            logger.info("Exploration interrupted (coverage not integrated)")
            self.stop_exploration()

        if self.config.exploration_limit and (uid+1) >= self.config.exploration_limit:
            logger.info('Exploration limit reached')
            self.stop_exploration()

        # Some analysis in post execution
        solve_time = self.seeds_manager.post_execution(execution, seed, not self._stop)
        self._total_emulation_time += expl_ts

        logger.info(f"Emulation: {self._fmt_secs(expl_ts)} | Solving: {self._fmt_secs(solve_time)} | Elapsed: {self._fmt_secs(self.__time_delta())}\n")

    def step(self) -> None:
        """
        Perform a single exploration step. That means it execute
        a single :py:obj:`SymbolicExecutor`. Then it gives the hand
        back to the user.
        """
        # Take an input
        seed = self.seeds_manager.pick_seed()

        # If we don't have any new seed to process just switch exploration to idle
        if seed is None:
            logger.info("worklist of seed to process is empty")
            self.status = ExplorationStatus.IDLE
            return

        # Iterate the callback to be called at each steps
        for cb in self.cbm.get_exploration_step_callbacks():
            cb(self)

        # Execution into a thread
        t = threading.Thread(
            name='\033[0;%dm[exec:%08d]\033[0m' % ((31 + (self.uid_counter % 4)), self.uid_counter),
            target=self._worker,
            args=[seed, self.uid_counter],
            daemon=True
        )
        t.start()
        self.uid_counter += 1

        while True:
            t.join(0.001)
            if not t.is_alive():
                break

    def explore(self) -> ExplorationStatus:
        """
        Start the symbolic exploration. That function
        holds until the exploration is interrupted or finished.

        :returns: the status of the exploration
        :rtype: ExplorationStatus
        """
        self.status = ExplorationStatus.RUNNING

        try:
            while self.seeds_manager.seeds_available() and not self._stop:
                gc.collect()
                self.step()

            if self.status == ExplorationStatus.RUNNING:
                if not self.seeds_manager.seeds_available():
                    logger.info("exploration step done (no new seed available)")
                    self.status = ExplorationStatus.IDLE
                else:
                    logger.warning(f'should not exit step() in RUNNING state (stop? {self._stop}, seeds available? {self.seeds_manager.seeds_available()})')

        except KeyboardInterrupt:
            logger.warning("keyboard interrupt, stop symbolic exploration")
            self.stop_exploration()

        self.post_exploration()
        logger.info(f"Total time of the exploration: {self._fmt_secs(self.__time_delta())}")

        if self.status == ExplorationStatus.IDLE:
            logger.info("Execution IDLE no seeds to execute")

        return self.status

    def add_input_seed(self, seed: Union[bytes, Seed]) -> None:
        """
        Add the given bytes or Seed object as input for the exploration.

        :param seed: input seed to add in the pending inputs to process
        :type seed: Union[bytes, Seed]
        """
        seed = seed if isinstance(seed, Seed) else Seed(seed)
        self.seeds_manager.add_new_seed(seed)

    def stop_exploration(self) -> None:
        """ Interrupt the exploration """
        self.status = ExplorationStatus.STOPPED
        self._stop = True

    def terminate_exploration(self) -> None:
        """ Terminate exploration with status terminated (normal shutdown) """
        self.status = ExplorationStatus.TERMINATED
        self._stop = True

    @staticmethod
    def _fmt_secs(seconds) -> str:
        m, s = divmod(seconds, 60)
        h, m = divmod(m, 60)
        return (f"{int(h)}h" if h else '')+f"{int(m)}m{int(s)}s"

    def post_exploration(self) -> None:
        """ Perform  all calls to post exploration functions"""
        self.seeds_manager.post_exploration()
        self.coverage.post_exploration(self.workspace)



# built-in imports
from __future__ import annotations
import shutil
from pathlib import Path
from typing import Generator, Optional, Union
import time

# local imports
from tritondse.types import PathLike
from tritondse.seed import Seed, SeedStatus
import tritondse.logging

logger = tritondse.logging.get()


class Workspace(object):
    """
    Class to abstract the file tree of the current exploration workspace.
    A user willing to save additional files in the workspace is invited
    to do it from the workspace API as it somehow abstract the exact
    location of it.
    """

    DEFAULT_WORKSPACE = "/tmp/triton_workspace"
    CORPUS_DIR = "corpus"
    CRASH_DIR = "crashes"
    HANG_DIR = "hangs"
    FAIL_DIR = "fails"
    WORKLIST_DIR = "worklist"
    METADATA_DIR = "metadata"
    BIN_DIR = "bin"
    LOG_FILE = "tritondse.log"

    def __init__(self, root_dir: PathLike):
        """
        :param root_dir: Root directory of the workspace. Created if not existing
        :type root_dir: :py:obj:`tritondse.types.PathLike`
        """
        if not root_dir:  # If no workspace was provided create a unique temporary one
            self.root_dir = Path(self.DEFAULT_WORKSPACE) / str(time.time()).replace(".", "")
            self.root_dir.mkdir(parents=True)
        else:
            self.root_dir = Path(root_dir)
            if not self.root_dir.exists():  # Create the directory in case it was not existing
                self.root_dir.mkdir(parents=True)
                self.initialize()
        self.root_dir: Path = self.root_dir.resolve()  #: root directory of the Workspace

    def initialize(self, flush: bool = False) -> None:
        """
        Initialize the workspace by creating all required subfolders
        if not already existing.

        :param flush: if True deletes all files contained in the workspace
        :type flush: bool
        """

        for dir in (self.root_dir / x for x in [self.CORPUS_DIR, self.CRASH_DIR, self.HANG_DIR, self.WORKLIST_DIR, self.METADATA_DIR, self.BIN_DIR, self.FAIL_DIR]):
            if not dir.exists():
                logger.debug(f"Creating the {dir} directory")
                dir.mkdir(parents=True)
            else:
                if flush:
                    shutil.rmtree(dir)
                    dir.mkdir()

    def get_metadata_file(self, name: str) -> Optional[str]:
        """
        Read a metadata file from the workspace on disk.
        Data is read as a string. If the given file does not
        exists, None is returned

        :param name: file name (can also be a path)
        :type name: str
        :returns: File content as string if existing
        :rtype: Optional[str]
        """
        p = (self.root_dir / self.METADATA_DIR) / name
        if p.exists():
            return p.read_text()
        else:
            return None

    def get_metadata_file_path(self, name: str) -> Path:
        """
        Get a file path in the workspace directory that the user
        can write into. Might be called for the user to write on
        its own the file content. If name is a file tree, all parent
        directories are created.

        :param name: filename wanted
        :type name: str
        :return: absolute filepath (regardless of whether it exists or not)
        """
        p = (self.root_dir / self.METADATA_DIR) / name
        if not p.parent.exists():
            p.parent.mkdir(parents=True)
        return p

    def get_binary_directory(self) -> Path:
        """
        Get the directory containing the executable (and its dependencies).
        :return: Path of the directory
        """
        return self.root_dir / self.BIN_DIR

    def save_metadata_file(self, name: str, content: Union[str, bytes]) -> None:
        """
        Save ``content`` in a file ``name`` in the metadata directory.
        The name should be a file name not a path.

        :param name: file name
        :type name: str
        :param content: content of the file to write
        :type content: Union[str, bytes]
        """
        p = (self.root_dir / self.METADATA_DIR) / name
        if isinstance(content, str):
            p.write_text(content)
        else:
            p.write_bytes(content)

    def _iter_seeds(self, directory: str, st: SeedStatus) -> Generator[Seed, None, None]:
        """ Iterate over seeds """
        for file in (self.root_dir/directory).glob("*.cov"):
            yield Seed.from_file(file, st)

    def iter_corpus(self) -> Generator[Seed, None, None]:
        """
        Iterate over the corpus files as Seed object.

        :returns: generator of Seed object
        :rtype: Generator[Seed, None, None]
        """
        yield from self._iter_seeds(self.CORPUS_DIR, SeedStatus.OK_DONE)

    def iter_crashes(self) -> Generator[Seed, None, None]:
        """
        Iterate over the crashes files as Seed object.

        :returns: generator of Seed object
        :rtype: Generator[Seed, None, None]
        """
        yield from self._iter_seeds(self.CRASH_DIR, SeedStatus.CRASH)

    def iter_hangs(self) -> Generator[Seed, None, None]:
        """
        Iterate over the hang files as Seed object.

        :returns: generator of Seed object
        :rtype: Generator[Seed, None, None]
        """
        yield from self._iter_seeds(self.HANG_DIR, SeedStatus.HANG)

    def iter_worklist(self) -> Generator[Seed, None, None]:
        """
        Iterate over the worklist files as Seed object.
        Worklist are all the pending seeds

        :returns: generator of Seed object
        :rtype: Generator[Seed, None, None]
        """
        yield from self._iter_seeds(self.WORKLIST_DIR, SeedStatus.NEW)

    def iter_fails(self) -> Generator[Seed, None, None]:
        """
        Iterate over the fail files as Seed object.

        :returns: generator of Seed object
        :rtype: Generator[Seed, None, None]
        """
        yield from self._iter_seeds(self.FAIL_DIR, SeedStatus.FAIL)

    def save_seed(self, seed: Seed) -> None:
        """
        Save the current seed in the workspace directory matching its status.

        :param seed: Seed to save
        :type seed: Seed
        """
        mapper = {SeedStatus.NEW: self.WORKLIST_DIR,
                  SeedStatus.OK_DONE: self.CORPUS_DIR,
                  SeedStatus.HANG: self.HANG_DIR,
                  SeedStatus.CRASH: self.CRASH_DIR,
                  SeedStatus.FAIL: self.FAIL_DIR}
        p = (self.root_dir / mapper[seed.status]) / seed.filename
        p.write_bytes(bytes(seed))

    def update_seed_location(self, seed: Seed) -> None:
        """
        Move a worklist seed to its final location according to its (new) status.
        Typically used to move a seed from pending ones to corpus or crash once it
        is fully consumed.

        :param seed: seed to move
        :type seed: Seed
        """
        old_p = (self.root_dir / self.WORKLIST_DIR) / seed.filename
        try:
            old_p.unlink()  # Remove the seed from the worklist
        except:
            logger.warning(f"seed {seed} unlink failed")
            pass  # FIXME: Not meant to get here
        self.save_seed(seed)

    def save_file(self, rel_path: PathLike, content: Union[str, bytes], override: bool = False):
        """
        Save a, arbitrary file in the workspace by providing the relative path
        of the file. If ``override`` is True, erase the previous file if any.

        :param rel_path: relative path of the file
        :type rel_path: :py:obj:`tritondse.types.PathLike`
        :param content: content to write
        :type content: Union[str, bytes]
        :param override: whether to override or not an existing file
        :type override: bool
        """
        p = self.root_dir / rel_path
        p.parent.mkdir(parents=True, exist_ok=True)
        if not p.exists() or override:
            if isinstance(content, str):
                p.write_text(content)
            elif isinstance(content, bytes):
                p.write_bytes(content)
            else:
                assert False

    @property
    def logfile_path(self):
        return self.root_dir / self.LOG_FILE
